"""
SMAT API access.
"""

from datetime import datetime

import requests


class SmatError(Exception):
    pass


class InvalidSiteError(SmatError):
    pass


class InvalidIntervalError(SmatError):
    pass


class FailedRequestError(SmatError):
    pass


class Smat:
    """
    Class to handle SMAT API calls.
    """

    def __init__(self):
        self.base_url = "https://api.smat-app.com"
        self.sites = [
            "telegram",
            "rumble_video",
            "rumble_comment",
            "bitchute_video",
            "bitchute_comment",
            "tiktok_video",
            "tiktok_comment",
            "rutube_video",
            "rutube_comment",
            "lbry_video",
            "lbry_comment",
            "8kun",
            "4chan",
            "gab",
            "parler",
            "win",
            "poal",
            "kiwifarms",
            "gettr",
            "wimkin",
            "mewe",
            "minds",
            "vk",
            "truth_social",
        ]
        self.intervals = [
            "hour",
            "day",
            "week",
            "month",
            "year",
        ]

    def content(
        self,
        term: str,
        limit: int,
        site: str,
        since: datetime,
        until: datetime,
        esquery: bool = True,
        sortdesc: bool = False,
    ) -> dict:
        """
        Query the /content endpoint of the SMAT API and return results as a dict of the JSON response.

        :param term: str
        :param limit: int
        :param site: str
        :param since: datetime.datetime
        :param until: datetime.datetime
        :param esquery: bool
        :param sortdesc: bool
        :return: dict
        """
        validated_site = self.validate_site(site)
        endpoint = "/content"
        response = requests.get(
            self.base_url + endpoint,
            params={
                "term": term,
                "limit": limit,
                "site": validated_site,
                "since": since,
                "until": until,
                "esquery": esquery,
                "sortdesc": sortdesc,
            },
        )
        return self._handle_response(response)

    def timeseries(
        self,
        term: str,
        interval: str,
        site: str,
        since: datetime,
        until: datetime,
        esquery: bool = True,
        changepoint: bool = False,
    ) -> dict:
        """
        Query the /timeseries endpoint of the SMAT API and return results as a dict of the JSON response. Note:
        currently can not do changepoints.

        :param term: str
        :param interval: str
        :param site: str
        :param since: datetime.datetime
        """
        # TODO: implement a way to do changepoints

        validated_site = self.validate_site(site)
        validated_interval = self.validate_interval(interval)
        endpoint = "/timeseries"
        response = requests.get(
            self.base_url + endpoint,
            params={
                "term": term,
                "interval": validated_interval,
                "site": validated_site,
                "since": since,
                "until": until,
                "esquery": esquery,
                "changepoint": changepoint,
            },
        )
        return self._handle_response(response)

    def activity(
        self,
        term: str,
        agg_by: str,
        site: str,
        since: datetime,
        until: datetime,
        esquery: bool = True,
    ) -> dict:
        """
        Query the /activity endpoint of the SMAT API and return results as a dict of the JSON response.

        :param term: str
        :param agg_by: str
        :param site: str
        :param since: datetime.datetime
        :param until: datetime.datetime
        :param esquery: bool
        :return: dict
        """
        validated_site = self.validate_site(site)
        endpoint = "/activity"
        response = requests.get(
            self.base_url + endpoint,
            params={
                "term": term,
                "agg_by": agg_by,
                "site": validated_site,
                "since": since,
                "until": until,
                "esquery": esquery,
            },
        )
        return self._handle_response(response)

    def validate_site(self, site: str) -> str:
        """
        Raises exception if the provided site does not exist in the list of valid sites, otherwise return the site
        provided.

        :param site: str
        :return: str
        """
        if site not in self.sites:
            raise InvalidSiteError(
                f"{site} is an invalid site. Valid sites are: {', '.join(sorted(self.sites))}"
            )
        return site

    def validate_interval(self, interval):
        """
        Raises exception if the provided interval does not exist in the list of valid intervals, otherwise return the
        interval provided.

        :param interval: str
        :return: str
        """
        if interval not in self.intervals:
            raise InvalidIntervalError(
                f"{interval} is an invalid interval. Valid intervals are: {', '.join(sorted(self.intervals))}"
            )
        return interval

    def _handle_response(self, response):
        if response.status_code == 200:
            return response.json()
        else:
            raise FailedRequestError(
                f"Failed request [{response.status_code}]: {response.text}"
            )



"""
Formatters for SMAT API responses.
"""
import json
from abc import ABC, abstractmethod
from typing import Iterable


class ResponseFormatter(ABC):
    """
    Abstract base class for a formatter.

    Subclasses should override content, timeseries, and activity methods and should return iterables of strings.
    """

    def __init__(self, data: dict) -> None:
        self.data = data

    @abstractmethod
    def content(self) -> Iterable[str]:
        raise NotImplementedError

    @abstractmethod
    def timeseries(self) -> Iterable[str]:
        raise NotImplementedError

    @abstractmethod
    def activity(self, agg_by) -> Iterable[str]:
        raise NotImplementedError


class NdjsonFormatter(ResponseFormatter):
    """
    Ndjson formatter.
    """

    def content(self) -> Iterable[str]:
        results = []
        for hit in self.data["hits"]["hits"]:
            results.append(json.dumps(hit["_source"]))
        return results

    def timeseries(self) -> Iterable[str]:
        results = []
        for hit in self.data["aggregations"]["date"]["buckets"]:
            results.append(json.dumps(hit))
        return results

    def activity(self, agg_by) -> Iterable[str]:
        results = []
        for hit in self.data["aggregations"][agg_by]["buckets"]:
            results.append(json.dumps(hit))
        return results


class JsonFormatter(ResponseFormatter):
    """
    Json formatter.
    """

    def content(self) -> Iterable[str]:
        results = []
        for hit in self.data["hits"]["hits"]:
            results.append(hit["_source"])
        return [json.dumps(dict(results=results))]

    def timeseries(self) -> Iterable[str]:
        results = []
        for hit in self.data["aggregations"]["date"]["buckets"]:
            results.append(hit)
        changepoint = self.data["aggregations"].get("changepoint", {})
        return [json.dumps(dict(results=results, changepoint=changepoint))]

    def activity(self, agg_by) -> Iterable[str]:
        results = []
        for hit in self.data["aggregations"][agg_by]["buckets"]:
            results.append(hit)
        return [json.dumps(dict(results=results))]



"""
CLI for the SMAT API.
"""

import datetime

import click

from smat_cli import Smat, SmatError
from smat_cli.formatters import JsonFormatter, NdjsonFormatter


def until_default():
    return (datetime.date.today() - datetime.timedelta(days=7)).strftime("%Y-%m-%d")


@click.group()
@click.option(
    "--format",
    "-f",
    default="ndjson",
    type=click.Choice(["ndjson", "json"]),
    help="Format to return data in.",
    show_default=True,
)
@click.pass_context
def main(ctx, format):
    """
    smat is a tool for querying the Social Media Analysis Toolkit (https://www.smat-app.com). It supports querying
    via the public API, which may be subject to rate limits, and responses may take substantial time if the server is
    busy. All commands can accept ElasticSearch query string syntax (see ElasticSearch documentation).

    \f
    This is the main Click group for the cli, accessible via the package's entrypoint "smat".
    """
    ctx.ensure_object(dict)
    formatters = {"ndjson": NdjsonFormatter, "json": JsonFormatter}
    ctx.obj["formatter"] = formatters[format]


@main.command()
@click.pass_context
@click.option(
    "--limit",
    "-l",
    default=10,
    type=int,
    help="Maximum number of results to return.",
    show_default=True,
)
@click.option(
    "--site",
    "-s",
    required=True,
    help="The site to get the content from.",
)
@click.option(
    "--since",
    default=until_default(),
    type=click.DateTime(),
    help="Earliest datetime of the content to return.",
    show_default=True,
)
@click.option(
    "--until",
    type=click.DateTime(),
    help="Latest datetime of the content to return.",
)
@click.argument("query", type=str)
def content(ctx, limit, site, since, until, query):
    """
    Get content matching a query from a site.

    \f
    A Click command for the main group that makes requests to the /content endpoint and outputs JSON results.
    """
    api = Smat()
    try:
        formatter = ctx.obj["formatter"](
            api.content(term=query, limit=limit, site=site, since=since, until=until)
        )
    except SmatError as e:
        click.echo("Error: " + str(e))

    for each in formatter.content():
        click.echo(each)


@main.command()
@click.pass_context
@click.option(
    "--site",
    "-s",
    required=True,
    help="The site to get the content from.",
)
@click.option(
    "--since",
    default=until_default(),
    type=click.DateTime(),
    help="Earliest datetime of the content to return.",
    show_default=True,
)
@click.option(
    "--until",
    type=click.DateTime(),
    help="Latest datetime of the content to return.",
)
@click.option(
    "--interval",
    "-i",
    type=str,
    help="Interval for the data being returned.",
)
@click.option(
    "--changepoint/--no-changepoint",
    type=bool,
    default=False,
    help="Include changepoint (only for json format).",
)
@click.argument("query", type=str)
def timeseries(ctx, query, interval, site, since, until, changepoint=False):
    """
    Get timeseries matching a query from a site.

    \f
    A Click command for the main group that makes requests to the /timeseries endpoint and outputs JSON results.
    """
    if not ctx.obj["formatter"] == JsonFormatter and changepoint:
        raise click.ClickException("`--changepoint` must be used with `--format json`")
    api = Smat()
    try:
        formatter = ctx.obj["formatter"](
            api.timeseries(
                term=query,
                since=since,
                until=until,
                interval=interval,
                site=site,
                changepoint=changepoint,
            )
        )
    except SmatError as e:
        click.echo("Error: " + str(e))

    for each in formatter.timeseries():
        click.echo(each)


@main.command()
@click.pass_context
@click.option(
    "--site",
    "-s",
    required=True,
    help="The site to get the content from.",
)
@click.option(
    "--since",
    default=until_default(),
    type=click.DateTime(),
    help="Earliest datetime of the content to return.",
    show_default=True,
)
@click.option(
    "--until",
    type=click.DateTime(),
    help="Latest datetime of the content to return.",
)
@click.option(
    "--agg-by",
    "-a",
    type=str,
    required=True,
    help="Response key to aggregate by.",
)
@click.argument("query", type=str)
def activity(ctx, query, site, since, until, agg_by):
    """
    Get activity aggregated by key matching a query from a site.

    The agg-by option for this command is the key to use for the aggregation and the possible options can vary depending
    on which site is being queried. For example, `--agg-by channelusername --site telegram trump` will aggregate all
    mentions of Trump on Telegram by channel. In order to better understand the available keys for aggregation, using
    the `content` tool can provide some examples.

    \f
    A Click command for the main group that makes requests to the /activity endpoint and outputs JSON results.
    """
    api = Smat()
    try:
        formatter = ctx.obj["formatter"](
            api.activity(
                term=query,
                since=since,
                until=until,
                site=site,
                agg_by=agg_by,
            )
        )
    except SmatError as e:
        click.echo("Error: " + str(e))

    for each in formatter.activity(agg_by):
        click.echo(each)


@main.command()
def sites():
    """
    Show a list of available sites.

    \f
    A Click command for the main group that outputs a list of sites available via the SMAT API.
    """
    api = Smat()
    for site in sorted(api.sites):
        click.echo(site)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['smat_cli']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.4,<9.0.0', 'requests>=2.27.1,<3.0.0', 'urllib3>=1.26.9,<2.0.0']

entry_points = \
{'console_scripts': ['smat = smat_cli.cli:main']}

setup_kwargs = {
    'name': 'smat-cli',
    'version': '0.1.3',
    'description': 'A command line tool and library to get data from the Social Media Analysis Toolkit (SMAT).',
    'long_description': '# SMAT-CLI\n> Provides command line tools for getting data from the Social Media Analysis Toolkit ([SMAT](https://www.smat-app.com))\n> as well as a library for interacting with SMAT from your own code.\n\n[![Python Versions][python-image]](https://)\n[![Latest Release][latest-release-image]](https://pypi.org/project/smat-cli/)\n[![Pipeline Status][pipeline-status-image]](https://gitlab.com/dhosterman/smat-cli)\n\nThe [Social Media Analysis Toolkit](https://www.smat-app.com) is a resource that allows activists, journalists, \nresearchers, and other social good organizations to collect information about hate, mis/disinformation, and extremism\nfrom a variety of online platforms. The folks at SMAT are providing an amazing service and deserve your support! Go to\ntheir [Open Collective](https://opencollective.com/smat) page to support them if you\'re able.\n\nSMAT-CLI is a tool that makes getting that information from the API easy, either from your terminal or as part of your\nown application.\n\n## Installation\n\n**OS X & Linux:**\n\n```sh\npip install smat-cli\n```\n\nThough, I recommend using [Pipx](https://github.com/pypa/pipx) to install it as a a system tool.\n\n```sh\npipx install smat-cli\n```\n\n**Windows:**\n\nComing soon!\n\n## Usage Examples\n\nLet\'s say you want to collect 1000 posts from Telegram posted between Jan 6 to March 1, 2021. You can do that with the \n`content` command like this.\n\n```sh\nsmat content -s telegram -l 1000 --since 2021-01-06 --until 2021-03-01 trump\n```\n\nIf you want some aggregated data, you can use the `timeseries` command to fetch a count of posts mentioning Trump from \nJan 6 to March 1, 2021 and aggregate those into daily buckets, you can use the following.\n\n```sh\nsmat timeseries -s telegram -i day --since 2021-01-06 --until 2021-03-01 trump\n```\n\nYou can also aggregate by any arbitrary key present in the data for the site. To get an idea of which keys are available,\nyou can examine the results of a `content` command. Once you know the key you want to aggregate on, you can use `activity`\nto, for example, count the number of posts containing the term Trump in each Telegram channel from Jan 6 to March 1, 2022.\n\n```sh\nsmat activity -s telegram -a channelusername --since 2021-01-06 --until 2021-03-01 trump\n```\n\nAll above commands print output line-by-line to stdout in JSON, so you can pipe the results to a file in the normal way.\n\n```sh\n\nsmat content -s telegram -l 1000 --since 2021-01-06 --until 2021-03-01 trump >> data.ndjson\n```\n\nYou can also specify different formats for the output. For example, if you\'d like the results in JSON instead of \nJSONlines, you can pass `--format json`. This is currently the only way to format changepoint data when using \n`timeseries`.\n\n```sh\nsmat --format json content -s telegram -l 1000 --since 2021-01-06 --until 2021-03-01 trump > data.json\n```\n\nIn addition, this package can be used in another application by importing `Smat` from `smat_cli` and using it to query\nthe API from inside your program.\n\n```python\nfrom smat_cli import Smat\napi = Smat()\ndata = api.content(term="trump", site="telegram", ...)\nfor d in data:\n    print(d["message"])\n```\n\n\n## Development Setup\n\nThis project uses [Poetry](https://python-poetry.org/) and the code is formatted with [Black](https://github.com/psf/black).\n\nTests can be run via [Tox](https://tox.wiki/en/latest/), which will run the tests in Python verions: 3.7, 3.8, 3.9, 3.10.\nFor this to work properly, all of these Python versions must be installed. A Dockerfile has been included in the `test_runner`\ndirectory if necessary.\n\n```sh\npoetry install\ntox\n```\n\n## Release History\n\n* 0.1.3\n  * Add new sites: TikTok, Rutube, Truth Social (@benzonip)\n  * Added support for Python 3.11\n* 0.1.2\n  * Updates to package info and README\n  * Add formatters and JsonFormatter\n* 0.1.1\n  * Updates to package info and README\n* 0.1.0\n  * Initial release\n\n## Me\n\nDaniel Hosterman – [@dhosterman](https://twitter.com/dhosterman) – daniel@danielhosterman.com\n\nDistributed under the [Unlicense](http://unlicense.org) license. See ``LICENSE`` for more information.\n\n[https://gitlab.com/dhosterman](https://gitlab.com/dhosterman/)\n\n## Contributing\n\n1. Fork it (<https://github.com/yourname/yourproject/fork>)\n2. Create your feature branch (`git checkout -b feature/fooBar`)\n3. Format your code with [Black](https://github.com/psf/black)\n4. Ensure there are tests for your changes and that they pass\n5. Commit your changes (`git commit -am \'Add some fooBar\'`)\n6. Push to the branch (`git push origin feature/fooBar`)\n7. Create a new Pull Request\n\n## Contributors\n\n* Peter Benzoni (@benzonip)\n\n<!-- Markdown link & img dfn\'s -->\n[python-image]: https://img.shields.io/badge/python-3.7%20%7C%203.8%20%7C%203.9%20%7C%203.10%20%7C%203.11-informational\n[latest-release-image]: https://img.shields.io/badge/latest%20release-0.1.2-informational\n[pipeline-status-image]: https://gitlab.com/dhosterman/smat-cli/badges/main/pipeline.svg\n',
    'author': 'dhosterman',
    'author_email': 'daniel@danielhosterman.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://gitlab.com/dhosterman/smat-cli',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
